from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _floating_pi,
    _is_negative_zero,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarSin(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarSin":
        return ScalarSin(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.sin, ScalarSin
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.sin(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and sin(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.sin(argv)

        # apply the inverse function to get the bounds on arg
        # ensure that the bounds on sin(...) are in [-1, +1]
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.asin(
            _maximum_zero_sign_sensitive(Xs.dtype.type(-1), expr_lower)
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.asin(
            _minimum_zero_sign_sensitive(expr_upper, Xs.dtype.type(1))
        )

        # sin(...) is periodic, so we need to drop to difference bounds before
        #  applying the difference to argv to stay in the same period
        argv_asin = np.asin(exprv)
        arg_lower_diff = arg_lower - argv_asin
        arg_upper_diff = arg_upper - argv_asin

        # np.asin maps to [-pi/2, +pi/2] where sin is monotonically increasing
        # flip the argument difference bounds where sin is monotonically
        #  decreasing
        # we check monotonicity using the derivative sin'(x) = cos(x)
        needs_flip = np.cos(argv) < 0

        # check for the case where any finite value would work
        full_domain: np.ndarray[tuple[Ps], np.dtype[np.bool]] = np.less_equal(
            expr_lower, -1
        ) & np.greater_equal(expr_upper, 1)

        fmax = np.finfo(Xs.dtype).max

        # sin(+-inf) = NaN, so force infinite argv to have exact bounds
        # sin(finite) in [-1, +1] so allow any finite argv if the all of
        #  [-1, +1] is allowed
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        # if expr_upper == -0.0 and arg_upper == 0, we need to guarantee that
        #  arg_upper is also -0.0
        # FIXME: how do we handle bounds right next to the peak where the
        #        expression bounds could be exceeded inside the interval?
        # TODO: the intervals can sometimes be extended if expr_lower <= -1 or
        #       expr_upper >= 1
        # TODO: since sin is periodic, an interval union could be used in the
        #       future
        arg_lower = _ensure_array(arg_lower_diff, copy=True)
        np.negative(arg_upper_diff, out=arg_lower, where=needs_flip)
        np.add(arg_lower, argv, out=arg_lower)
        arg_lower[full_domain] = -fmax
        np.copyto(arg_lower, argv, where=np.isinf(argv), casting="no")
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper = _ensure_array(arg_upper_diff, copy=True)
        np.negative(arg_lower_diff, out=arg_upper, where=needs_flip)
        np.add(arg_upper, argv, out=arg_upper)
        arg_upper[full_domain] = fmax
        np.copyto(arg_upper, argv, where=np.isinf(argv), casting="no")
        arg_upper[(arg_upper == 0) & _is_negative_zero(expr_upper)] = -0.0
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in sin(asin(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.sin(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.sin(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"sin({self._a!r})"


class ScalarCos(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarCos":
        return ScalarCos(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.cos, ScalarCos
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.cos(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and cos(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.cos(argv)

        # apply the inverse function to get the bounds on arg
        # ensure that the bounds on cos(...) are in [-1, +1]
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.acos(
            _maximum_zero_sign_sensitive(Xs.dtype.type(-1), expr_lower)
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.acos(
            _minimum_zero_sign_sensitive(expr_upper, Xs.dtype.type(1))
        )

        # cos(...) is periodic, so we need to drop to difference bounds before
        #  applying the difference to argv to stay in the same period
        argv_acos = np.acos(exprv)
        arg_lower_diff = argv_acos - arg_lower
        arg_upper_diff = argv_acos - arg_upper

        # np.acos maps to [pi, 0] where cos is monotonically decreasing
        # flip the argument difference bounds where cos is monotonically
        #  decreasing
        # we check monotonicity using the derivative cos'(x) = -sin(x)
        needs_flip = np.sin(argv) >= 0

        # check for the case where any finite value would work
        full_domain: np.ndarray[tuple[Ps], np.dtype[np.bool]] = np.less_equal(
            expr_lower, -1
        ) & np.greater_equal(expr_upper, 1)

        fmax = np.finfo(Xs.dtype).max

        # cps(+-inf) = NaN, so force infinite argv to have exact bounds
        # cos(finite) in [-1, +1] so allow any finite argv if the all of
        #  [-1, +1] is allowed
        # FIXME: how do we handle bounds right next to the peak where the
        #        expression bounds could be exceeded inside the interval?
        # TODO: the intervals can sometimes be extended if expr_lower <= -1 or
        #       expr_upper >= 1
        # TODO: since cos is periodic, an interval union could be used in the
        #       future
        arg_lower = _ensure_array(arg_lower_diff, copy=True)
        np.negative(arg_upper_diff, out=arg_lower, where=needs_flip)
        np.add(arg_lower, argv, out=arg_lower)
        arg_lower[full_domain] = -fmax
        np.copyto(arg_lower, argv, where=np.isinf(argv), casting="no")
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper = _ensure_array(arg_upper_diff, copy=True)
        np.negative(arg_lower_diff, out=arg_upper, where=needs_flip)
        np.add(arg_upper, argv, out=arg_upper)
        arg_upper[full_domain] = fmax
        np.copyto(arg_upper, argv, where=np.isinf(argv), casting="no")
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in cos(acos(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.cos(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.cos(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"cos({self._a!r})"


class ScalarTan(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarTan":
        return ScalarTan(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.tan, ScalarTan
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.tan(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and tan(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.tan(argv)

        # apply the inverse function to get the bounds on arg
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.atan(expr_lower)
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.atan(expr_upper)

        # tan(...) is periodic, so we need to drop to difference bounds before
        #  applying the difference to argv to stay in the same period
        argv_atan = np.atan(exprv)
        arg_lower_diff = arg_lower - argv_atan
        arg_upper_diff = arg_upper - argv_atan

        # check for the case where any finite value would work
        full_domain: np.ndarray[tuple[Ps], np.dtype[np.bool]] = (
            expr_lower == Xs.dtype.type(-np.inf)
        ) & (expr_upper == Xs.dtype.type(np.inf))

        fmax = np.finfo(Xs.dtype).max

        # tan(+-inf) = NaN, so force infinite argv to have exact bounds
        # tan(finite) in [-inf, +inf] so allow any finite argv if the all of
        #  [-inf, +inf] is allowed
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        # if expr_upper == -0.0 and arg_upper == 0, we need to guarantee that
        #  arg_upper is also -0.0
        # FIXME: how do we handle bounds right next to the peak where the
        #        expression bounds could be exceeded inside the interval?
        # TODO: since tan is periodic, an interval union could be used in the
        #       future
        arg_lower = _ensure_array(np.add(argv, arg_lower_diff))
        arg_lower[full_domain] = -fmax
        np.copyto(arg_lower, argv, where=np.isinf(argv), casting="no")
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper = _ensure_array(np.add(argv, arg_upper_diff))
        arg_upper[full_domain] = fmax
        np.copyto(arg_upper, argv, where=np.isinf(argv), casting="no")
        arg_upper[(arg_upper == 0) & _is_negative_zero(expr_upper)] = -0.0
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in tan(atan(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.tan(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.tan(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"tan({self._a!r})"


class ScalarAsin(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAsin":
        return ScalarAsin(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.asin, ScalarAsin
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.asin(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and asin(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.asin(argv)

        pi = _floating_pi(Xs.dtype)
        pi_2: F = np.divide(pi, 2)
        one_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(2))

        # apply the inverse function to get the bounds on arg
        # asin(...) is NaN when abs(...) > 1 and can then take any value > 1
        # otherwise ensure that the bounds on asin(...) are in [-pi/2, +pi/2]
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.sin(_maximum_zero_sign_sensitive(-pi_2, expr_lower))
        )
        arg_lower[np.greater(argv, 1)] = one_eps
        arg_lower[np.less(argv, -1)] = -np.inf
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.sin(_minimum_zero_sign_sensitive(expr_upper, pi_2))
        )
        arg_upper[np.greater(argv, 1)] = np.inf
        arg_upper[np.less(argv, -1)] = -one_eps
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper and abs(argv) < 1
        np.copyto(
            arg_lower,
            argv,
            where=((expr_lower == expr_upper) & np.less(np.abs(argv), 1)),
            casting="no",
        )
        np.copyto(
            arg_upper,
            argv,
            where=((expr_lower == expr_upper) & np.less(np.abs(argv), 1)),
            casting="no",
        )

        # handle rounding errors in asin(sin(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.asin(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.asin(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"asin({self._a!r})"


class ScalarAcos(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAcos":
        return ScalarAcos(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.acos, ScalarAcos
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.acos(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and acos(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.acos(argv)

        pi = _floating_pi(Xs.dtype)
        one_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(2))

        # apply the inverse function to get the bounds on arg
        # acos(...) is NaN when abs(...) > 1 and can then take any value > 1
        # otherwise ensure that the bounds on acos(...) are in [0, pi]
        # since cos is monotonically decreasing in [0, pi], the lower and upper
        #  bounds get switched
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.cos(_minimum_zero_sign_sensitive(expr_upper, pi))
        )
        arg_lower[np.greater(argv, 1)] = one_eps
        arg_lower[np.less(argv, -1)] = -np.inf
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.cos(_maximum_zero_sign_sensitive(Xs.dtype.type(0), expr_lower))
        )
        arg_upper[np.greater(argv, 1)] = np.inf
        arg_upper[np.less(argv, -1)] = -one_eps
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper and abs(argv) < 1
        np.copyto(
            arg_lower,
            argv,
            where=((expr_lower == expr_upper) & np.less(np.abs(argv), 1)),
            casting="no",
        )
        np.copyto(
            arg_upper,
            argv,
            where=((expr_lower == expr_upper) & np.less(np.abs(argv), 1)),
            casting="no",
        )

        # handle rounding errors in acos(cos(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.acos(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.acos(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"acos({self._a!r})"


class ScalarAtan(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAtan":
        return ScalarAtan(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.atan, ScalarAtan
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.atan(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and atan(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.atan(argv)

        # compute pi/2 but guard against rounding error to ensure that
        #  tan(pi/2) >> 0
        atan_max: F = np.atan(Xs.dtype.type(np.inf))
        if np.tan(atan_max) < 0:
            atan_max = Xs.dtype.type(np.nextafter(atan_max, Xs.dtype.type(0)))
        assert np.tan(atan_max) > 0

        # apply the inverse function to get the bounds on arg
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        # ensure that the bounds on atan(...) are in [-pi/2, +pi/2]
        # since tan is discontinuous at +-pi/2, we need to be extra careful
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.tan(expr_lower)
        )
        arg_lower[expr_lower < -atan_max] = -np.inf
        arg_lower[expr_lower > atan_max] = np.inf
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.tan(expr_upper)
        )
        arg_upper[expr_upper < -atan_max] = -np.inf
        arg_upper[expr_upper > atan_max] = np.inf
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in atan(tan(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.atan(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.atan(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"atan({self._a!r})"



from collections.abc import Callable, Mapping
from typing import overload

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Fi, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarEqual(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarEqual":
        return ScalarEqual(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: comparison_to_dtype(np.equal, a, b, dtype),
            ScalarEqual,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return comparison_to_dtype(
            lambda a, b: a == b,
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b

        a_const = not a.has_data
        b_const = not b.has_data
        assert not (a_const and b_const), "constant == has no data bounds"

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        bv_nxt_av = _ensure_array(np.nextafter(bv, av))

        # compute a finite midpoint for a and b
        #  - if either is NaN, we won't use this midpoint
        #  - since we use finite values, mid(-Inf, +Inf) = 0
        av_mid_bv: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(  # type: ignore
            np.divide(np.nan_to_num(av), Xs.dtype.type(2)),
            np.divide(np.nan_to_num(bv), Xs.dtype.type(2)),
        )
        # the midpoint must be in range [av, bv), unless av == bv, since we
        #  later want to be able to step from the midpoint closer towards bv
        av_mid_bv = _maximum_zero_sign_sensitive(
            av_mid_bv, _minimum_zero_sign_sensitive(av, bv_nxt_av)
        )
        av_mid_bv = _minimum_zero_sign_sensitive(
            av_mid_bv, _maximum_zero_sign_sensitive(av, bv_nxt_av)
        )

        # compute the next value from b towards a that can be part of a's
        #  safe interval
        # if b is const this is the literal next value
        # otherwise it is the midpoint
        if not b_const:
            np.copyto(
                bv_nxt_av,
                av_mid_bv,
                where=b.eval_has_data(Xs, late_bound),
                casting="no",
            )

        # compute the next value from a towards b that can be part of b's
        #  safe interval
        # if a is const this is the literal next value
        # otherwise it is the next value from the midpoint towards b
        av_nxt_bv = _ensure_array(av, copy=True)
        if not a_const:
            np.copyto(
                av_nxt_bv,
                av_mid_bv,
                where=a.eval_has_data(Xs, late_bound),
                casting="no",
            )
        av_nxt_bv = np.nextafter(av_nxt_bv, bv)

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, (a == b) = True, then
        #  - keep av and bv the same
        #  - we do *not* special-case -0.0 and +0.0 here
        # if expr_upper < 1, (a == b) = False, then
        #  - if av > bv, restrict a's lower and b's upper bound to not overlap
        #  - if av < bv, restrict a's upper and b's lower bound to not overlap
        #  - otherwise av or bv is NaN and any non-NaN argument can be anything
        # otherwise, (a == b) in [True, False] and all args can be anything
        if not a_const:
            a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(a_lower, av, where=np.greater(expr_lower, 0), casting="no")
            np.copyto(
                a_lower,
                bv_nxt_av,
                where=(np.less(expr_upper, 1) & (av > bv)),
                casting="no",
            )

            a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(a_upper, av, where=np.greater(expr_lower, 0), casting="no")
            np.copyto(
                a_upper,
                bv_nxt_av,
                where=(np.less(expr_upper, 1) & (av < bv)),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, a_lower))
            a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, a_upper))

            # recurse into arg a
            Xs_lower_, Xs_upper_ = a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        # cont. with b as outlined above
        if not b_const:
            b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(b_lower, bv, where=np.greater(expr_lower, 0), casting="no")
            np.copyto(
                b_lower,
                av_nxt_bv,
                where=(np.less(expr_upper, 1) & (bv > av)),
                casting="no",
            )

            b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(b_upper, bv, where=np.greater(expr_lower, 0), casting="no")
            np.copyto(
                b_upper,
                av_nxt_bv,
                where=(np.less(expr_upper, 1) & (bv < av)),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
            b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

            # recurse into arg b
            xl, xu = b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} == {self._b!r}"


class ScalarNotEqual(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarNotEqual":
        return ScalarNotEqual(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: comparison_to_dtype(np.not_equal, a, b, dtype),
            ScalarNotEqual,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return comparison_to_dtype(
            lambda a, b: a != b,
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b

        a_const = not a.has_data
        b_const = not b.has_data
        assert not (a_const and b_const), "constant != has no data bounds"

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        bv_nxt_av = _ensure_array(np.nextafter(bv, av))

        # compute a finite midpoint for a and b
        #  - if either is NaN, we won't use this midpoint
        #  - since we use finite values, mid(-Inf, +Inf) = 0
        av_mid_bv: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(  # type: ignore
            np.divide(np.nan_to_num(av), Xs.dtype.type(2)),
            np.divide(np.nan_to_num(bv), Xs.dtype.type(2)),
        )
        # the midpoint must be in range [av, bv), unless av == bv, since we
        #  later want to be able to step from the midpoint closer towards bv
        av_mid_bv = _maximum_zero_sign_sensitive(
            av_mid_bv, _minimum_zero_sign_sensitive(av, bv_nxt_av)
        )
        av_mid_bv = _minimum_zero_sign_sensitive(
            av_mid_bv, _maximum_zero_sign_sensitive(av, bv_nxt_av)
        )

        # compute the next value from b towards a that can be part of a's
        #  safe interval
        # if b is const this is the literal next value
        # otherwise it is the midpoint
        if not b_const:
            np.copyto(
                bv_nxt_av,
                av_mid_bv,
                where=b.eval_has_data(Xs, late_bound),
                casting="no",
            )

        # compute the next value from a towards b that can be part of b's
        #  safe interval
        # if a is const this is the literal next value
        # otherwise it is the next value from the midpoint towards b
        av_nxt_bv = _ensure_array(av, copy=True)
        if not a_const:
            np.copyto(
                av_nxt_bv,
                av_mid_bv,
                where=a.eval_has_data(Xs, late_bound),
                casting="no",
            )
        av_nxt_bv = np.nextafter(av_nxt_bv, bv)

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, (a != b) = True, then
        #  - if av > bv, restrict a's lower and b's upper bound to not overlap
        #  - if av < bv, restrict a's upper and b's lower bound to not overlap
        #  - otherwise av or bv is NaN and any non-NaN argument can be anything
        # if expr_upper < 1, (a != b) = False, then
        #  - keep av and bv the same
        #  - we do *not* special-case -0.0 and +0.0 here
        # otherwise, (a != b) in [True, False] and all args can be anything
        if not a_const:
            a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                a_lower,
                bv_nxt_av,
                where=(np.greater(expr_lower, 0) & (av > bv)),
                casting="no",
            )
            np.copyto(a_lower, av, where=np.less(expr_upper, 1), casting="no")

            a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                a_upper,
                bv_nxt_av,
                where=(np.greater(expr_lower, 0) & (av < bv)),
                casting="no",
            )
            np.copyto(a_upper, av, where=np.less(expr_upper, 1), casting="no")

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, a_lower))
            a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, a_upper))

            # recurse into arg a
            Xs_lower_, Xs_upper_ = a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        # cont. with b as outlined above
        if not b_const:
            b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                b_lower,
                av_nxt_bv,
                where=(np.greater(expr_lower, 0) & (bv > av)),
                casting="no",
            )
            np.copyto(b_lower, bv, where=np.less(expr_upper, 1), casting="no")

            b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                b_upper,
                av_nxt_bv,
                where=(np.greater(expr_lower, 0) & (bv < av)),
                casting="no",
            )
            np.copyto(b_upper, bv, where=np.less(expr_upper, 1), casting="no")

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
            b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

            # recurse into arg b
            xl, xu = b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} != {self._b!r}"


class ScalarLess(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarLess":
        return ScalarLess(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: comparison_to_dtype(np.less, a, b, dtype),
            ScalarLess,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return comparison_to_dtype(
            lambda a, b: a < b,  # type: ignore
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b

        a_const = not a.has_data
        b_const = not b.has_data
        assert not (a_const and b_const), "constant < has no data bounds"

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        bv_nxt_av = _ensure_array(np.nextafter(bv, av))

        # compute a finite midpoint for a and b
        #  - if either is NaN, we won't use this midpoint
        #  - since we use finite values, mid(-Inf, +Inf) = 0
        av_mid_bv: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(  # type: ignore
            np.divide(np.nan_to_num(av), Xs.dtype.type(2)),
            np.divide(np.nan_to_num(bv), Xs.dtype.type(2)),
        )
        # the midpoint must be in range [av, bv), unless av == bv, since we
        #  later want to be able to step from the midpoint closer towards bv
        av_mid_bv = _maximum_zero_sign_sensitive(
            av_mid_bv, _minimum_zero_sign_sensitive(av, bv_nxt_av)
        )
        av_mid_bv = _minimum_zero_sign_sensitive(
            av_mid_bv, _maximum_zero_sign_sensitive(av, bv_nxt_av)
        )

        # compute the next value from b towards a that can be part of a's
        #  safe interval
        # if b is const this is the literal next value
        # otherwise it is the midpoint
        if not b_const:
            np.copyto(
                bv_nxt_av,
                av_mid_bv,
                where=b.eval_has_data(Xs, late_bound),
                casting="no",
            )

        # compute the next value from a towards b that can be part of b's
        #  safe interval
        # if a is const this is the literal next value
        # otherwise it is the next value from the midpoint towards b
        av_nxt_bv = _ensure_array(av, copy=True)
        if not a_const:
            np.copyto(
                av_nxt_bv,
                av_mid_bv,
                where=a.eval_has_data(Xs, late_bound),
                casting="no",
            )
        av_nxt_bv = np.nextafter(av_nxt_bv, bv)

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, (a < b) = True, then
        #  - restrict a's upper and b's lower bound to not overlap
        # if expr_upper < 1, (a < b) = False, then
        #  - if av >= bv, restrict a's lower and b's upper bound
        #  - otherwise av or bv is NaN and any non-NaN argument can be anything
        # otherwise, (a < b) in [True, False] and all args can be anything
        if not a_const:
            a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                a_lower,
                bv_nxt_av,  # a >= b and bv_nxt_av >= b and >= mid_b
                where=(np.less(expr_upper, 1) & (av >= bv)),
                casting="no",
            )

            a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                a_upper,
                bv_nxt_av,  # a < b and bv_nxt_av < b and < mid_b
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, a_lower))
            a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, a_upper))

            # recurse into arg a
            Xs_lower_, Xs_upper_ = a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        # cont. with b as outlined above
        if not b_const:
            b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                b_lower,
                av_nxt_bv,  # a < b and av_nxt_bv > a and > mid_a
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                b_upper,
                av_nxt_bv,  # a >= b and av_nxt_bv <= a and <= mid_a
                where=(np.less(expr_upper, 1) & (av >= bv)),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
            b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

            # recurse into arg b
            xl, xu = b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} < {self._b!r}"


class ScalarGreaterEqual(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarGreaterEqual":
        return ScalarGreaterEqual(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: comparison_to_dtype(np.greater_equal, a, b, dtype),
            ScalarGreaterEqual,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return comparison_to_dtype(
            lambda a, b: a >= b,  # type: ignore
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b

        a_const = not a.has_data
        b_const = not b.has_data
        assert not (a_const and b_const), "constant >= has no data bounds"

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        bv_nxt_av = _ensure_array(np.nextafter(bv, av))

        # compute a finite midpoint for a and b
        #  - if either is NaN, we won't use this midpoint
        #  - since we use finite values, mid(-Inf, +Inf) = 0
        av_mid_bv: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(  # type: ignore
            np.divide(np.nan_to_num(av), Xs.dtype.type(2)),
            np.divide(np.nan_to_num(bv), Xs.dtype.type(2)),
        )
        # the midpoint must be in range [av, bv), unless av == bv, since we
        #  later want to be able to step from the midpoint closer towards bv
        av_mid_bv = _maximum_zero_sign_sensitive(
            av_mid_bv, _minimum_zero_sign_sensitive(av, bv_nxt_av)
        )
        av_mid_bv = _minimum_zero_sign_sensitive(
            av_mid_bv, _maximum_zero_sign_sensitive(av, bv_nxt_av)
        )

        # compute the next value from b towards a that can be part of a's
        #  safe interval
        # if b is const this is the literal next value
        # otherwise it is the midpoint
        if not b_const:
            np.copyto(
                bv_nxt_av,
                av_mid_bv,
                where=b.eval_has_data(Xs, late_bound),
                casting="no",
            )

        # compute the next value from a towards b that can be part of b's
        #  safe interval
        # if a is const this is the literal next value
        # otherwise it is the next value from the midpoint towards b
        av_nxt_bv = _ensure_array(av, copy=True)
        if not a_const:
            np.copyto(
                av_nxt_bv,
                av_mid_bv,
                where=a.eval_has_data(Xs, late_bound),
                casting="no",
            )
        av_nxt_bv = np.nextafter(av_nxt_bv, bv)

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, (a >= b) = True, then
        #  - restrict a's lower and b's upper bound
        # if expr_upper < 1, (a >= b) = False, then
        #  - if av < bv, restrict a's upper and b's lower bound to not overlap
        #  - otherwise av or bv is NaN and any non-NaN argument can be anything
        # otherwise, (a >= b) in [True, False] and all args can be anything
        if not a_const:
            a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                a_lower,
                bv_nxt_av,  # a >= b and bv_nxt_av >= b and >= mid_b
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                a_upper,
                bv_nxt_av,  # a < b and bv_nxt_av < b and < mid_b
                where=(np.less(expr_upper, 1) & (av < bv)),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, a_lower))
            a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, a_upper))

            # recurse into arg a
            Xs_lower_, Xs_upper_ = a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        # cont. with b as outlined above
        if not b_const:
            b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                b_lower,
                av_nxt_bv,  # a < b and av_nxt_bv > a and > mid_a
                where=(np.less(expr_upper, 1) & (av < bv)),
                casting="no",
            )

            b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                b_upper,
                av_nxt_bv,  # a >= b and av_nxt_bv <= a and <= mid_a
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
            b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

            # recurse into arg b
            xl, xu = b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} >= {self._b!r}"


class ScalarLessEqual(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarLessEqual":
        return ScalarLessEqual(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: comparison_to_dtype(np.less_equal, a, b, dtype),
            ScalarLessEqual,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return comparison_to_dtype(
            lambda a, b: a <= b,  # type: ignore
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b

        a_const = not a.has_data
        b_const = not b.has_data
        assert not (a_const and b_const), "constant <= has no data bounds"

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        av_nxt_bv = _ensure_array(np.nextafter(av, bv))

        # compute a finite midpoint for a and b
        #  - if either is NaN, we won't use this midpoint
        #  - since we use finite values, mid(-Inf, +Inf) = 0
        av_mid_bv: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(  # type: ignore
            np.divide(np.nan_to_num(av), Xs.dtype.type(2)),
            np.divide(np.nan_to_num(bv), Xs.dtype.type(2)),
        )
        # the midpoint must be in range (av, bv], unless av == bv, since we
        #  later want to be able to step from the midpoint closer towards av
        av_mid_bv = _maximum_zero_sign_sensitive(
            av_mid_bv, _minimum_zero_sign_sensitive(av_nxt_bv, bv)
        )
        av_mid_bv = _minimum_zero_sign_sensitive(
            av_mid_bv, _maximum_zero_sign_sensitive(av_nxt_bv, bv)
        )

        # compute the next value from b towards a that can be part of a's
        #  safe interval
        # if b is const this is the literal next value
        # otherwise it is the next value from the midpoint towards a
        bv_nxt_av = _ensure_array(bv, copy=True)
        if not b_const:
            np.copyto(
                bv_nxt_av,
                av_mid_bv,
                where=b.eval_has_data(Xs, late_bound),
                casting="no",
            )
        bv_nxt_av = np.nextafter(bv_nxt_av, av)

        # compute the next value from a towards b that can be part of b's
        #  safe interval
        # if a is const this is the literal next value
        # otherwise it is the midpoint
        if not a_const:
            np.copyto(
                av_nxt_bv,
                av_mid_bv,
                where=a.eval_has_data(Xs, late_bound),
                casting="no",
            )

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, (a <= b) = True, then
        #  - restrict a's upper and b's lower bound
        # if expr_upper < 1, (a <= b) = False, then
        #  - if av > bv, restrict a's lower and b's upper bound to not overlap
        #  - otherwise av or bv is NaN and any non-NaN argument can be anything
        # otherwise, (a <= b) in [True, False] and all args can be anything
        if not a_const:
            a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                a_lower,
                bv_nxt_av,  # a > b and bv_nxt_av > b and > mid_b
                where=(np.less(expr_upper, 1) & (av > bv)),
                casting="no",
            )

            a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                a_upper,
                bv_nxt_av,  # a <= b and bv_nxt_av <= b and <= mid_b
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, a_lower))
            a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, a_upper))

            # recurse into arg a
            Xs_lower_, Xs_upper_ = a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        # cont. with b as outlined above
        if not b_const:
            b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                b_lower,
                av_nxt_bv,  # a <= b and av_nxt_bv >= a and >= mid_a
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                b_upper,
                av_nxt_bv,  # a > b and av_nxt_bv < a and < mid_a
                where=(np.less(expr_upper, 1) & (av > bv)),
                casting="no",
            )

            b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
            b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            # recurse into arg b
            xl, xu = b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} <= {self._b!r}"


class ScalarGreater(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarGreater":
        return ScalarGreater(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: comparison_to_dtype(np.greater, a, b, dtype),
            ScalarGreater,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return comparison_to_dtype(
            lambda a, b: a > b,  # type: ignore
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b

        a_const = not a.has_data
        b_const = not b.has_data
        assert not (a_const and b_const), "constant > has no data bounds"

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        av_nxt_bv = _ensure_array(np.nextafter(av, bv))

        # compute a finite midpoint for a and b
        #  - if either is NaN, we won't use this midpoint
        #  - since we use finite values, mid(-Inf, +Inf) = 0
        av_mid_bv: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(  # type: ignore
            np.divide(np.nan_to_num(av), Xs.dtype.type(2)),
            np.divide(np.nan_to_num(bv), Xs.dtype.type(2)),
        )
        # the midpoint must be in range (av, bv], unless av == bv, since we
        #  later want to be able to step from the midpoint closer towards av
        av_mid_bv = _maximum_zero_sign_sensitive(
            av_mid_bv, _minimum_zero_sign_sensitive(av_nxt_bv, bv)
        )
        av_mid_bv = _minimum_zero_sign_sensitive(
            av_mid_bv, _maximum_zero_sign_sensitive(av_nxt_bv, bv)
        )

        # compute the next value from b towards a that can be part of a's
        #  safe interval
        # if b is const this is the literal next value
        # otherwise it is the next value from the midpoint towards a
        bv_nxt_av = _ensure_array(bv, copy=True)
        if not b_const:
            np.copyto(
                bv_nxt_av,
                av_mid_bv,
                where=b.eval_has_data(Xs, late_bound),
                casting="no",
            )
        bv_nxt_av = np.nextafter(bv_nxt_av, av)

        # compute the next value from a towards b that can be part of b's
        #  safe interval
        # if a is const this is the literal next value
        # otherwise it is the midpoint
        if not a_const:
            np.copyto(
                av_nxt_bv,
                av_mid_bv,
                where=a.eval_has_data(Xs, late_bound),
                casting="no",
            )

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, (a > b) = True, then
        #  - restrict a's lower and b's upper bound to not overlap
        # if expr_upper < 1, (a > b) = False, then
        #  - if av <= bv, restrict a's upper and b's lower bound
        #  - otherwise av or bv is NaN and any non-NaN argument can be anything
        # otherwise, (a > b) in [True, False] and all args can be anything
        if not a_const:
            a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                a_lower,
                bv_nxt_av,  # a > b and bv_nxt_av > b and > mid_b
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                a_upper,
                bv_nxt_av,  # a <= b and bv_nxt_av <= b and <= mid_b
                where=(np.less(expr_upper, 1) & (av <= bv)),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, a_lower))
            a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, a_upper))

            # recurse into arg a
            Xs_lower_, Xs_upper_ = a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        # cont. with b as outlined above
        if not b_const:
            b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            np.copyto(
                b_lower,
                av_nxt_bv,  # a <= b and av_nxt_bv >= a and >= mid_a
                where=(np.less(expr_upper, 1) & (av <= bv)),
                casting="no",
            )

            b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            np.copyto(
                b_upper,
                av_nxt_bv,  # a > b and av_nxt_bv < a and < mid_a
                where=np.greater(expr_lower, 0),
                casting="no",
            )

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
            b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

            # recurse into arg b
            xl, xu = b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} > {self._b!r}"


@overload
def comparison_to_dtype(
    classify: Callable[
        [np.ndarray[tuple[Ps], np.dtype[F]], np.ndarray[tuple[Ps], np.dtype[F]]],
        np.ndarray[tuple[Ps], np.dtype[np.bool]],
    ],
    a: np.ndarray[tuple[Ps], np.dtype[F]],
    b: np.ndarray[tuple[Ps], np.dtype[F]],
    dtype: np.dtype[F],
) -> np.ndarray[tuple[Ps], np.dtype[F]]: ...


@overload
def comparison_to_dtype(
    classify: Callable[[Fi, Fi], bool],
    a: Fi,
    b: Fi,
    dtype: np.dtype[Fi],
) -> Fi: ...


def comparison_to_dtype(comparison, a, b, dtype):
    c = comparison(a, b)

    if not isinstance(c, np.ndarray):
        return _ensure_array(c).astype(dtype, casting="safe")[()]

    return c.astype(dtype, casting="safe")



import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING, Generic, Self, TypeAlias, assert_never, final
from warnings import warn

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
    _zeros,
)
from ....utils.bindings import Parameter
from ....utils.error import QuantityOfInterestRuntimeWarning
from ..bound import DataBounds, data_bounds_checks, guarantee_data_within_expr_bounds
from ..typing import Es, F, Ns, Ps, np_sndarray

if TYPE_CHECKING:
    from .literal import Number


class Expr(ABC, Generic[*Es]):
    """
    Abstract base class for the quantity of interest expression abstract syntax
    tree.
    """

    __slots__: tuple[str, ...] = ()

    @property
    @abstractmethod
    def args(self) -> tuple[*Es]:
        """
        The sub-expression arguments of this expression.
        """

    @abstractmethod
    def with_args(self, *args: *Es) -> "Self | Number":
        """
        Reconstruct this expression with different sub-expression arguments.

        Parameters
        ----------
        *args : *Es
            The modified sub-expression arguments, derived from
            [`self.args`][..args].

        Returns
        -------
        expr : Self
            The modified expression.
        """

    @final
    def map_expr(self, m: "Callable[[AnyExpr], AnyExpr]") -> "AnyExpr":
        """
        Recursively maps the expression mapping function `m` over this
        expression and its sub-expression arguments.

        Parameters
        ----------
        m : Callable[[AnyExpr], AnyExpr]
            The expression mapper, which is applied to an expression whose
            sub-expression arguments have already been mapped, i.e. the mapper
            is *not* responsible for recursion.

        Returns
        -------
        expr : AnyExpr
            The mapped expression.
        """

        args: tuple[AnyExpr, ...] = self.args  # type: ignore
        mapped_args: tuple[AnyExpr, ...] = tuple(a.map_expr(m) for a in args)
        mapped_self: AnyExpr = self.with_args(*mapped_args)  # type: ignore
        return m(mapped_self)

    @final
    @property
    def expr_size(self) -> int:
        """
        The size of the expression tree, counting the number of nodes.
        """

        args: tuple[AnyExpr, ...] = self.args  # type: ignore

        return sum(a.expr_size for a in args) + 1

    @final
    @property
    def data_expr_size(self) -> int:
        """
        The size of the expression tree, counting the number of data-dependent
        nodes.
        """

        args: tuple[AnyExpr, ...] = self.args  # type: ignore

        if len(args) == 0:
            return int(self.has_data)

        return sum(a.expr_size for a in args if a.has_data) + 1

    @final
    @property
    def has_data(self) -> bool:
        """
        Does this expression reference the data `x` or `X[i]`?
        """

        args: tuple[AnyExpr, ...] = self.args  # type: ignore

        return any(a.has_data for a in args)

    @final
    def eval_has_data(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[np.bool]]:
        """
        Evaluate if the pointwise expression values references the data `x` or
        `X[i]`.

        Parameters
        ----------
        Xs : np_sndarray[Ps, Ns, np.dtype[F]]
            The stencil-extended data, in floating-point format, which must be
            of shape [Ps, ...stencil_shape].
        late_bound : Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]]
            The late-bound constants parameters for this expression, with the
            same shape and floating-point dtype as the stencil-extended data.

        Returns
        -------
        vals : np.ndarray[tuple[Ps], np.dtype[np.bool]]
            Pointwise [`True`][True] if the point references the data.
        """

        if not self.has_data:
            return _zeros((Xs.shape[0],), dtype=np.dtype(np.bool))

        args: tuple[AnyExpr, ...] = self.args  # type: ignore

        match args:
            case ():
                return _zeros((Xs.shape[0],), dtype=np.dtype(np.bool))
            case _:
                a, *as_ = args
                has_data = a.eval_has_data(Xs, late_bound)
                for a in as_:
                    has_data |= a.eval_has_data(Xs, late_bound)
                return has_data

    @final
    @property
    def data_indices(self) -> frozenset[tuple[int, ...]]:
        """
        The set of data indices `X[is]` that this expression uses.
        """

        args: tuple[AnyExpr, ...] = self.args  # type: ignore

        match args:
            case ():
                return frozenset()
            case (a,):
                return a.data_indices
            case _:
                indices: set[tuple[int, ...]] = set()
                for a in args:
                    indices.update(a.data_indices)
                return frozenset(indices)

    @final
    def apply_array_element_offset(
        self,
        axis: int,
        offset: int,
    ) -> "AnyExpr":
        """
        Apply an `offset` to the array element indices along the given `axis`.

        This method applies to data and late-bound constants.

        Parameters
        ----------
        axis : int
            The axis along which the array element indices are offset.
        offset : int
            The offset that is applied to the array element indices.

        Returns
        -------
        expr : AnyExpr
            The modified expression.
        """

        return self.with_args(
            *(a.apply_array_element_offset(axis, offset) for a in self.args)  # type: ignore
        )

    @final
    @property
    def late_bound_constants(self) -> frozenset[Parameter]:
        """
        The set of late-bound constant parameters that this expression uses.
        """

        args: tuple[AnyExpr, ...] = self.args  # type: ignore

        match args:
            case ():
                return frozenset()
            case (a,):
                return a.late_bound_constants
            case _:
                late_bound: set[Parameter] = set()
                for a in args:
                    late_bound.update(a.late_bound_constants)
                return frozenset(late_bound)

    # FIXME: constant_fold based on self.args and self.with_args is blocked on
    #        not being able to relate on TypeVarTuple to another, here *Expr to
    #        *F, see e.g. https://github.com/python/typing/issues/1216
    @abstractmethod
    def constant_fold(self, dtype: np.dtype[F]) -> "F | AnyExpr":
        """
        Apply scalar constant folding for the given `dtype` to this expression.

        Parameters
        ----------
        dtype : np.dtype[F]
            The dtype for which constant expressions are evaluated and folded.

        Returns
        -------
        folded : F | Expr
            The constant-folded scalar number or expression, which may have
            been simplified.
        """

    @abstractmethod
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        """
        Evaluate this expression on the stencil-extended data `Xs`.

        Parameters
        ----------
        Xs : np_sndarray[Ps, Ns, np.dtype[F]]
            The stencil-extended data, in floating-point format, which must be
            of shape [Ps, ...stencil_shape].
        late_bound : Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]]
            The late-bound constants parameters for this expression, with the
            same shape and floating-point dtype as the stencil-extended data.

        Returns
        -------
        vals : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise expression values.
        """

    @abstractmethod
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        """
        Compute the lower-upper bounds on the stencil-extended data `Xs` that
        that satisfy the lower-upper bounds `expr_lower` and `expr_lower` on
        this expression.

        This method should *not* be called manually.

        This method is allowed to return slightly wrongly-rounded results
        that are then corrected by
        [`compute_data_bounds`][..compute_data_bounds].

        If this method is known to have no rounding errors and always return
        the correct data bounds, it can be decorated with
        [`@data_bounds(DataBounds.infallible)`][.....bound.data_bounds].

        Parameters
        ----------
        expr_lower : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise lower bound on this expression.
        expr_upper : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise upper bound on this expression.
        Xs : np_sndarray[Ps, Ns, np.dtype[F]]
            The stencil-extended data, in floating-point format, which must be
            of shape [Ps, ...stencil_shape].
        late_bound : Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]]
            The late-bound constants parameters for this expression, with the
            same shape and floating-point dtype as the stencil-extended data.

        Returns
        -------
        Xs_lower, Xs_upper : tuple[np_sndarray[Ps, Ns, np.dtype[F]], np_sndarray[Ps, Ns, np.dtype[F]]]
            The stencil-extended lower and upper bounds on the stencil-extended
            data `Xs`.

            The bounds have not yet been combined across neighbouring points
            that contribute to the same QoI points.
        """

    @final
    def compute_data_bounds(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        """
        Compute the lower-upper bounds on the stencil-extended data `Xs` that
        that satisfy the lower-upper bounds `expr_lower` and `expr_lower` on
        this expression.

        This method, by default, calls into
        [`compute_data_bounds_unchecked`][..compute_data_bounds_unchecked]
        and then applies extensive rounding checks to ensure that the returned
        bounds satisfy the bounds on this expression.

        Parameters
        ----------
        expr_lower : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise lower bound on this expression.
        expr_upper : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise upper bound on this expression.
        Xs : np_sndarray[Ps, Ns, np.dtype[F]]
            The stencil-extended data, in floating-point format, which must be
            of shape [Ps, ...stencil_shape].
        late_bound : Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]]
            The late-bound constants parameters for this expression, with the
            same shape and floating-point dtype as the stencil-extended data.

        Returns
        -------
        Xs_lower, Xs_upper : tuple[np_sndarray[Ps, Ns, np.dtype[F]], np_sndarray[Ps, Ns, np.dtype[F]]]
            The stencil-extended lower and upper bounds on the stencil-extended
            data `Xs`.

            The bounds have not yet been combined across neighbouring points
            that contribute to the same QoI points.
        """

        if (
            data_bounds_checks(self.compute_data_bounds_unchecked)
            != DataBounds.infallible
        ):
            exprv: np.ndarray[tuple[Ps], np.dtype[F]] = self.eval(Xs, late_bound)
            if not np.all((expr_lower <= exprv) | np.isnan(exprv)):
                warn(
                    "expression lower bounds are above the expression values",
                    category=QuantityOfInterestRuntimeWarning,
                )
            if not np.all((expr_upper >= exprv) | np.isnan(exprv)):
                warn(
                    "expression upper bounds are below the expression values",
                    category=QuantityOfInterestRuntimeWarning,
                )
        else:
            exprv = None  # type: ignore

        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower, Xs_upper = self.compute_data_bounds_unchecked(
            expr_lower, expr_upper, Xs, late_bound
        )

        warn_on_bounds_exceeded: bool

        match data_bounds_checked := data_bounds_checks(
            self.compute_data_bounds_unchecked
        ):
            case DataBounds.infallible:
                return Xs_lower, Xs_upper
            case DataBounds.unchecked:
                warn_on_bounds_exceeded = False
            case DataBounds.checked:
                warn_on_bounds_exceeded = True
            case _:
                assert_never(data_bounds_checked)

        # ensure that the original data values are within the data bounds
        Xs_lower = _minimum_zero_sign_sensitive(Xs, Xs_lower)
        Xs_upper = _maximum_zero_sign_sensitive(Xs, Xs_upper)

        # handle rounding errors in the lower bound computation
        Xs_lower = guarantee_data_within_expr_bounds(
            lambda Xs_lower: self.eval(
                Xs_lower,
                late_bound,
            ),
            exprv,
            Xs,
            Xs_lower,
            expr_lower,
            expr_upper,
            warn_on_bounds_exceeded=warn_on_bounds_exceeded,
        )
        Xs_upper = guarantee_data_within_expr_bounds(
            lambda Xs_upper: self.eval(
                Xs_upper,
                late_bound,
            ),
            exprv,
            Xs,
            Xs_upper,
            expr_lower,
            expr_upper,
            warn_on_bounds_exceeded=warn_on_bounds_exceeded,
        )

        return Xs_lower, Xs_upper

    @abstractmethod
    @override
    def __repr__(self) -> str:
        pass


AnyExpr: TypeAlias = Expr[*tuple["AnyExpr", ...]]
""" Expression with sub-expression arguments """

if sys.version_info >= (3, 11) or TYPE_CHECKING:
    EmptyExpr: TypeAlias = Expr[()]
    """ Expression with zero arguments """
else:
    EmptyExpr: TypeAlias = Expr



from collections.abc import Callable, Mapping
from typing import overload

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
    _stack,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Fi, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarNot(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarNot":
        return ScalarNot(a)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            lambda x: combine_to_dtype(np.logical_not, x, dtype),
            ScalarNot,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return combine_to_dtype(np.logical_not, self._a.eval(Xs, late_bound), Xs.dtype)

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg
        arg = self._a
        argv = arg.eval(Xs, late_bound)

        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, not(arg) = True, so arg must stay False, i.e zero
        # if expr_upper < 1, not(arg) = False, arg must stay True, i.e. non-zero
        #  only one of the two disjoint non-zero intervals is propagated
        # otherwise, not(arg) in [True, False] and arg can be anything
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
            Xs.shape[:1], Xs.dtype.type(-np.inf)
        )
        arg_lower[np.greater(expr_lower, 0)] = -0.0
        arg_lower[np.less(expr_upper, 1) & np.greater(argv, 0)] = smallest_subnormal

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
            Xs.shape[:1], Xs.dtype.type(np.inf)
        )
        arg_upper[np.greater(expr_lower, 0)] = +0.0
        arg_upper[np.less(expr_upper, 1) & np.less(argv, 0)] = -smallest_subnormal

        # TODO: an interval union could represent that the two disjoint
        #       intervals in the future

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"not({self._a!r})"


class ScalarAll(Expr[AnyExpr, AnyExpr, *tuple[AnyExpr, ...]]):
    __slots__: tuple[str, ...] = ("_a", "_b", "_cs")
    _a: AnyExpr
    _b: AnyExpr
    _cs: tuple[AnyExpr, ...]

    def __init__(self, a: AnyExpr, b: AnyExpr, *cs: AnyExpr):
        self._a = a
        self._b = b
        self._cs = cs

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr, *tuple[AnyExpr, ...]]:
        return (self._a, self._b, *self._cs)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr, *cs: AnyExpr) -> "ScalarAll":
        return ScalarAll(a, b, *cs)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        # we first individually fold each term
        fs = [abc.constant_fold(dtype) for abc in (self._a, self._b, *self._cs)]

        # any False -> False
        if any(f == 0 for f in fs):
            return dtype.type(0)

        # only keep non-const terms, which filters out all True-thy terms
        es = [f for f in fs if isinstance(f, Expr)]

        match es:
            case []:
                # all True-thy -> True
                return dtype.type(1)
            case [e]:
                # single non-const -> propagate
                return e
            case es:
                # all combinator over non-const terms
                a, b, *cs = es
                return ScalarAll(a, b, *cs)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return reduce_combine_to_dtype(
            lambda a: np.all(a, axis=0),
            _stack(
                [self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound)]
                + [c.eval(Xs, late_bound) for c in self._cs]
            ),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b
        cs = self._cs

        a_const = not a.has_data
        b_const = not b.has_data
        cs_const = [not c.has_data for c in cs]
        assert not (a_const and b_const and all(cs_const)), (
            "constant all has no data bounds"
        )

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)
        cvs = [c.eval(Xs, late_bound) for c in cs]

        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        a_const_zero = (av == 0) & (
            True if a_const else ~a.eval_has_data(Xs, late_bound)
        )
        b_const_zero = (bv == 0) & (
            True if b_const else ~b.eval_has_data(Xs, late_bound)
        )
        cs_const_zero = [
            (cv == 0) & (True if c_const else ~c.eval_has_data(Xs, late_bound))
            for c_const, c, cv in zip(cs_const, cs, cvs)
        ]

        any_constant_zero = _ensure_array(a_const_zero, copy=True)
        any_constant_zero |= b_const_zero
        for c_const_zero in cs_const_zero:
            any_constant_zero |= c_const_zero

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, all(*args) = True, then
        #  all args must stay True, i.e. non-zero
        # if expr_upper < 1, all(*args) = False, then
        #  - if any constant term is zero, all non-constant args can be anything
        #  - otherwise
        #    - all False (zero) args must stay False, i.e. zero
        #    - all True (non-zero) args can be anything
        # only one of the two disjoint non-zero intervals is propagated
        # otherwise, all(*args) in [True, False] and all args can be anything
        for term, tv, t_const_zero, t_const in zip(
            (a, b, *cs),
            (av, bv, *cvs),
            (a_const_zero, b_const_zero, *cs_const_zero),
            (a_const, b_const, *cs_const),
        ):
            if t_const:
                continue

            term_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            term_lower[np.greater(expr_lower, 0) & np.greater(tv, 0)] = (
                smallest_subnormal
            )
            term_lower[
                np.less(expr_upper, 1) & (tv == 0) & (~any_constant_zero | t_const_zero)
            ] = -0.0

            term_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            term_upper[np.greater(expr_lower, 0) & np.less(tv, 0)] = -smallest_subnormal
            term_upper[
                np.less(expr_upper, 1) & (tv == 0) & (~any_constant_zero | t_const_zero)
            ] = +0.0

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            # recurse into the terms
            xl, xu = term.compute_data_bounds(
                term_lower,
                term_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        abc = ", ".join([repr(self._a), repr(self._b)] + [repr(c) for c in self._cs])
        return f"all({abc})"


class ScalarAny(Expr[AnyExpr, AnyExpr, *tuple[AnyExpr, ...]]):
    __slots__: tuple[str, ...] = ("_a", "_b", "_cs")
    _a: AnyExpr
    _b: AnyExpr
    _cs: tuple[AnyExpr, ...]

    def __init__(self, a: AnyExpr, b: AnyExpr, *cs: AnyExpr):
        self._a = a
        self._b = b
        self._cs = cs

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr, *tuple[AnyExpr, ...]]:
        return (self._a, self._b, *self._cs)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr, *cs: AnyExpr) -> "ScalarAny":
        return ScalarAny(a, b, *cs)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        # we first individually fold each term
        fs = [abc.constant_fold(dtype) for abc in (self._a, self._b, *self._cs)]

        # any True -> True
        if any((not isinstance(f, Expr) and (f != 0)) for f in fs):
            return dtype.type(1)

        # only keep non-const terms, which filters out all False terms
        es = [f for f in fs if isinstance(f, Expr)]

        match es:
            case []:
                # all False -> False
                return dtype.type(0)
            case [e]:
                # single non-const -> propagate
                return e
            case es:
                # all combinator over non-const terms
                a, b, *cs = es
                return ScalarAny(a, b, *cs)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return reduce_combine_to_dtype(
            lambda a: np.any(a, axis=0),
            _stack(
                [self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound)]
                + [c.eval(Xs, late_bound) for c in self._cs]
            ),
            Xs.dtype,
        )

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a = self._a
        b = self._b
        cs = self._cs

        a_const = not a.has_data
        b_const = not b.has_data
        cs_const = [not c.has_data for c in cs]
        assert not (a_const and b_const and all(cs_const)), (
            "constant any has no data bounds"
        )

        # evaluate args
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)
        cvs = [c.eval(Xs, late_bound) for c in cs]

        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        a_const_non_zero = (av != 0) & (
            True if a_const else ~a.eval_has_data(Xs, late_bound)
        )
        b_const_non_zero = (bv != 0) & (
            True if b_const else ~b.eval_has_data(Xs, late_bound)
        )
        cs_const_non_zero = [
            (cv != 0) & (True if c_const else ~c.eval_has_data(Xs, late_bound))
            for c_const, c, cv in zip(cs_const, cs, cvs)
        ]

        any_constant_non_zero = _ensure_array(a_const_non_zero, copy=True)
        any_constant_non_zero |= b_const_non_zero
        for c_const_zero in cs_const_non_zero:
            any_constant_non_zero |= c_const_zero

        xl: np_sndarray[Ps, Ns, np.dtype[F]]
        xu: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
        Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, any(*args) = True, then
        #  - if any constant term is non-zero, all non-constant args can be
        #    anything
        #  - otherwise
        #    - all False (zero) args can be anything
        #    - all True (non-zero) args must stay True, i.e. non-zero
        # only one of the two disjoint non-zero intervals is propagated
        # if expr_upper < 1, any(*args) = False, then
        #  all args must stay False, i.e. zero
        # otherwise, any(*args) in [True, False] and all args can be anything
        for term, tv, t_const_non_zero, t_const in zip(
            (a, b, *cs),
            (av, bv, *cvs),
            (a_const_non_zero, b_const_non_zero, *cs_const_non_zero),
            (a_const, b_const, *cs_const),
        ):
            if t_const:
                continue

            term_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(-np.inf)
            )
            term_lower[
                np.greater(expr_lower, 0)
                & (tv != 0)
                & (~any_constant_non_zero | t_const_non_zero)
                & np.greater(tv, 0)
            ] = smallest_subnormal
            term_lower[np.less(expr_upper, 1)] = -0.0

            term_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                Xs.shape[:1], Xs.dtype.type(np.inf)
            )
            term_upper[
                np.greater(expr_lower, 0)
                & (tv != 0)
                & (~any_constant_non_zero | t_const_non_zero)
                & np.less(tv, 0)
            ] = -smallest_subnormal
            term_upper[np.less(expr_upper, 1)] = +0.0

            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future

            # recurse into the terms
            xl, xu = term.compute_data_bounds(
                term_lower,
                term_upper,
                Xs,
                late_bound,
            )

            # combine the inner data bounds
            if Xs_lower_ is None:
                Xs_lower_ = xl
            else:
                Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
            if Xs_upper_ is None:
                Xs_upper_ = xu
            else:
                Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        assert Xs_lower_ is not None
        assert Xs_upper_ is not None
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        abc = ", ".join([repr(self._a), repr(self._b)] + [repr(c) for c in self._cs])
        return f"any({abc})"


@overload
def combine_to_dtype(
    combine: Callable[
        [np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[np.bool]]
    ],
    a: np.ndarray[tuple[Ps], np.dtype[F]],
    dtype: np.dtype[F],
) -> np.ndarray[tuple[Ps], np.dtype[F]]: ...


@overload
def combine_to_dtype(
    combine: Callable[[Fi], bool],
    a: Fi,
    dtype: np.dtype[Fi],
) -> Fi: ...


def combine_to_dtype(combine, a, dtype):
    c = combine(a)

    if not isinstance(c, np.ndarray):
        return _ensure_array(c).astype(dtype, casting="safe")[()]

    return c.astype(dtype, casting="safe")


def reduce_combine_to_dtype(
    reduce_combine: Callable[
        [np.ndarray[tuple[int, Ps], np.dtype[F]]],
        np.ndarray[tuple[Ps], np.dtype[np.bool]],
    ],
    a: np.ndarray[tuple[int, Ps], np.dtype[F]],
    dtype: np.dtype[F],
):
    return reduce_combine(a).astype(dtype, casting="safe")



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils.bindings import Parameter
from ..bound import DataBounds, data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .literal import Number


class Group(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_expr",)
    _expr: AnyExpr

    def __new__(cls, expr: AnyExpr) -> "Group | Number":  # type: ignore[misc]
        if isinstance(expr, Number | Group):
            return expr
        # we have to also assign inside __new__ since the double Group
        #  unwrapping would still call __init__ and create a reference loop
        this = super().__new__(cls)
        this._expr = expr
        return this

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._expr,)

    @override
    def with_args(self, expr: AnyExpr) -> "Group | Number":
        return Group(expr)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        fexpr = self._expr.constant_fold(dtype)
        # partially / not constant folded -> stop further folding
        if isinstance(fexpr, Expr):
            return Group(fexpr)
        # fully constant folded -> allow further folding
        return fexpr

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return self._expr.eval(Xs, late_bound)

    @data_bounds(DataBounds.infallible)
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        return self._expr.compute_data_bounds(expr_lower, expr_upper, Xs, late_bound)

    @override
    def __repr__(self) -> str:
        return f"({self._expr!r})"
